;===============================================================================
; Implementation of AES (Rijndael) 128 bit on PIC16F877                        ;
;===============================================================================
; This implementation contains both of encryption and decryption routine already :)
; It implement full keyed (buffered) of GF(256) and SBOX.
; To unroll unnecessary loops and functions I combined processes such 
; add round key, shift row and substitution.
;
; Encryption Benchmark:
; Plain Text 	= 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
; Key		= 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
; Cipher Text	= 66 e9 4b d4 ef 8a 2c 3b 88 4c fa 59 ca 34 2b 2e
;
;
; Decryption Benchmark:
; Cipher Text 	= 66 e9 4b d4 ef 8a 2c 3b 88 4c fa 59 ca 34 2b 2e
; Key		= 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
; Plain Text	= 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
;
; Initially, data placed on s0:s15. The Encryption process encrypts the values in the
; memory s0:s15 and put the cipher on the same address, the same doe the decryption 
; process.
; Since decryption process go backward we have to make sure that we started with the 
; last subkey first and the "lastk" routine shoul come first before making directly
; decrypt the data, just to make sure we dealt with that.
; 
; But if we do the encryption and decryption sequentially, never mind with last
; subkey issue, since decrytion will use the last subkey that generated by
; encryption routine.
;
;
; Memory Mapping:
;  - s0:s15 are 16 bytes of input
;  - k0:k15 are 16 bytes of key





	LIST	P=PIC16F876
	INCLUDE P16F876.INC
	__CONFIG B'11111111110001'
	
;===============================================================================
; SFR Definition
;===============================================================================
status	equ	0x03
pcl	equ	0x02
pclath	equ	0x0a
fsr	equ	0x04
indf	equ	0x00
;===============================================================================
; Bit Definition
;===============================================================================
z	equ	0x02
c	equ	0x00
w	equ	0x00
f	equ	0x01
rp0	equ	0x05
rd	equ	0x00
;===============================================================================
; GPR Definition
;===============================================================================
tmp		equ	0x20
tmp1	equ	0x21
tmp2	equ	0x22
tmp3	equ	0x23
tmpc	equ	0x24
lcnt	equ	0x25
kcnt	equ	0x26
d0		equ	0x27
d1		equ	0x28
d2		equ	0x29
d3		equ	0x2a
;===============================================================================
;Plain and Cipher text bytes memory map
;===============================================================================
s0	equ	0x30
s1	equ	0x31
s2	equ	0x32
s3	equ	0x33
s4	equ	0x34
s5	equ	0x35
s6	equ	0x36
s7	equ	0x37
s8	equ	0x38
s9	equ	0x39
s10	equ	0x3a
s11	equ	0x3b
s12	equ	0x3c
s13	equ	0x3d
s14	equ	0x3e
s15	equ	0x3f
;===============================================================================
;Key bytes memory map
;===============================================================================
k0	equ	0x40	; Key Organized as follows
k1	equ	0x41	;
k2	equ	0x42	; k0  k1  k2  k3
k3	equ	0x43    ; k4  k5  k6  k7
k4	equ	0x44    ; k8  k9  k10 k11
k5	equ	0x45    ; k12 k13 k14 k15
k6	equ	0x46
k7	equ	0x47
k8	equ	0x48
k9	equ	0x49
k10	equ	0x4a
k11	equ	0x4b
k12	equ	0x4c
k13	equ	0x4d
k14	equ	0x4e
k15	equ	0x4f
;===============================================================================
; Macro to XOR 2 registers
;===============================================================================
xor2	MACRO	dst,arg1,arg2
	movf	arg2,w
	xorwf	arg1,w
	movwf	dst
	ENDM
;===============================================================================
; Macro to XOR 3 registers
;===============================================================================
xor3	MACRO	dst,arg1,arg2,arg3
	movf	arg3,w
	xorwf	arg2,w
	xorwf	arg1,w
	movwf	dst
	ENDM
;===============================================================================
; Macro to XOR 4 registers
;===============================================================================
xor4	MACRO	dst,arg1,arg2,arg3,arg4
	movf	arg4,w
	xorwf	arg3,w
	xorwf	arg2,w
	xorwf	arg1,w
	movwf	dst
	ENDM
;===============================================================================
; Macro to perform bytewise XOR
;===============================================================================
xorff	MACRO	a,b
	movf	b,w
	xorwf	a,f
	ENDM
;===============================================================================
; Macro to perform registy to registry copy
;===============================================================================
movff	MACRO	a,b
	movf	b,w
	movwf	a
	ENDM
;===============================================================================
; Macro to perform registy to registry copy including substitution
;===============================================================================
movfcf	MACRO	a,b
	movf	b,w
	call	lookup
	movwf	a
	ENDM
;===============================================================================
; Macro to load a registry by an immediate value
;===============================================================================
movlf	MACRO	c,d
	movlw	d
	movwf	c
	ENDM

;===============================================================================
; Macro to obtain 2*x in GF(2^8)
;===============================================================================
get2	MACRO	arg
	bcf	status,c
	rlf	arg,w
	btfsc	arg,7
	xorlw	0x1b
	movwf	arg
	ENDM

;===============================================================================
; Macro to obtain (2*x)+y in GF(2^8)
;===============================================================================
get2a	MACRO	arg1,arg2
	bcf	status,c
	rlf	arg1,w
	btfsc	arg1,7
	xorlw	0x1b
	xorwf	arg2,w
	movwf	arg1
	ENDM
;===============================================================================
;Macro for subbox
;===============================================================================
SubBox MACRO a
	movf a,w
	call lookup
	movwf a
	ENDM

;===============================================================================
; Program Main Entrance
;===============================================================================
	org	0x00

;Start by initializing the key and the text to be encrypted.
init_kp call initplain
		call initkey

	
; Start encryption Routine
encr	movlf	kcnt,0x8d
		movlf	lcnt,0x09
		call  	addrndk
encra	call  	fSubByte
		call  	fShiftRow
		call  	mixcol
		call	getk
		call 	addrndk
		decfsz	lcnt,f
		goto	encra
		call 	fSubByte
		call 	fShiftRow
		call	getk
		call  	addrndk
		nop	; encryption ends here

decr	movlf	lcnt,0x09
		call	addrndk
decra	call	iShiftRow
		call	iSubByte
		call 	igetk
		call	addrndk
		call	imixcol
		clrf	pclath
		decfsz	lcnt,f
		goto	decra
		call 	iShiftRow
		call	iSubByte
		call 	igetk
		call	addrndk
		nop		; decryption ends here
		goto init_kp
	

;==============================================================================
;Routine to initialize the key value.
;==============================================================================
initkey movlf k0,0x00
		movlf k1,0x01
		movlf k2,0x02
		movlf k3,0x03
		movlf k4,0x04
		movlf k5,0x05
		movlf k6,0x06
		movlf k7,0x07
		movlf k8,0x08
		movlf k9,0x09
		movlf k10,0x0a
		movlf k11,0x0b
		movlf k12,0x0c
		movlf k13,0x0d
		movlf k14,0x0e
		movlf k15,0x0f
		return

;==============================================================================
;Routine to initialize the plain text value
;==============================================================================	
initplain movlf s0,0x00
		  movlf s1,0x01
		  movlf s2,0x02
		  movlf s3,0x03
		  movlf s4,0x04
		  movlf s5,0x05
		  movlf s6,0x06
		  movlf s7,0x07
		  movlf s8,0x08
		  movlf s9,0x09
		  movlf s10,0x0a
		  movlf s11,0x0b
		  movlf s12,0x0c
		  movlf s13,0x0d
		  movlf s14,0x0e
		  movlf s15,0x0f
		  return
;===============================================================================
; Routine to get last subkey
;===============================================================================
lastk	movlf	lcnt,0x0a
lastka	call	getk
	decfsz	lcnt,f
	goto	lastka
	return
;===============================================================================
; Routine to Process Add Round Key
;===============================================================================
addrndk	xorff	s0, k0
	xorff	s1, k1
	xorff	s2, k2
	xorff	s3, k3
	xorff	s4, k4
	xorff	s5, k5
	xorff	s6, k6
	xorff	s7, k7
	xorff	s8, k8
	xorff	s9, k9
	xorff	s10,k10
	xorff	s11,k11
	xorff	s12,k12
	xorff	s13,k13
	xorff	s14,k14
	xorff	s15,k15
	return
;================================================================================
;Routine that performs forward subbyte
;================================================================================
fSubByte  movlf pclath, sbox
	call SubByte
	return
;================================================================================
;Routine that performs inverse subbyte
;================================================================================
iSubByte movlf pclath, isbox
	call SubByte
	return
;================================================================================
;Generic routine for lookup
;================================================================================
SubByte		SubBox s0
	SubBox s1
	SubBox s2
	SubBox s3
	SubBox s4
	SubBox s5
	SubBox s6
	SubBox s7
	SubBox s8
	SubBox s9
	SubBox s10
	SubBox s11
	SubBox s12
	SubBox s13
	SubBox s14
	SubBox s15
	return
;===============================================================================
;Routine that performs shiftrow
;===============================================================================
fShiftRow  movff tmp, s1
	movff s1,s5
	movff s5,s9
	movff s9,s13
	movff s13,tmp

	movff tmp,s2
	movff s2,s10
	movff s10,tmp
	movff tmp,s6
	movff s6,s14
	movff s14,tmp

	movff tmp,s15
	movff s15,s11
	movff s11,s7
	movff s7,s3
	movff s3,tmp
	return
;===============================================================================
;Routine that performs inverse shiftrow
;===============================================================================
iShiftRow  movff tmp, s13
	movff s13,s9
	movff s9,s5
	movff s5,s1
	movff s1,tmp

	movff tmp,s2
	movff s2,s10
	movff s10,tmp
	movff tmp,s6
	movff s6,s14
	movff s14,tmp

	movff tmp,s3
	movff s3,s7
	movff s7,s11
	movff s11,s15
	movff s15,tmp
	return
;===============================================================================
; Function to generate next subkey
;===============================================================================
getk	movlf	pclath,sbox
	movf	k13,w
	call	lookup
	xorwf	k0,f
	movf	k14,w
	call	lookup
	xorwf	k1,f
	movf	k15,w
	call	lookup
	xorwf	k2,f
	movf	k12,w
	call	lookup
	xorwf	k3,f

	get2	kcnt
	xorwf	k0,f

	xorff	k4, k0
	xorff	k5, k1
	xorff	k6, k2
	xorff	k7, k3
	xorff	k8, k4
	xorff	k9, k5
	xorff	k10,k6
	xorff	k11,k7
	xorff	k12,k8
	xorff	k13,k9
	xorff	k14,k10
	xorff	k15,k11	
	return
;==============================================================================
; Function to generate next subkey
;==============================================================================
igetk	xorff	k15,k11
	xorff	k14,k10
	xorff	k13,k9
	xorff	k12,k8
	xorff	k11,k7
	xorff	k10,k6
	xorff	k9,k5
	xorff	k8,k4
	xorff	k7,k3
	xorff	k6,k2
	xorff	k5,k1
	xorff	k4,k0

	movlf	pclath,sbox
	movf	k13,w
	call	lookup
	xorwf	kcnt,w
	xorwf	k0,f
	movf	k14,w
	call	lookup
	xorwf	k1,f
	movf	k15,w
	call	lookup
	xorwf	k2,f
	movf	k12,w
	call	lookup
	xorwf	k3,f

	bcf	status,c
	rrf	kcnt,w
	btfsc	status,c
	movlw	0x80
	movwf	kcnt
	return
;===============================================================================
; Subroutine to perform mix column function
;===============================================================================
; |                 |    |         |
; | s0  s1  s2  s3  |    | 2 1 1 3 |
; | s4  s5  s6  s7  |    | 3 2 1 1 |
; | s8  s9  s10 s11 |    | 1 3 2 1 |
; | s12 s13 s14 s15 |    | 1 1 3 2 |
; |                 |    |         |

mixcol	movf	s0,w	; row 1
	movwf	tmp1
	xorwf	s1,w
	xorwf	s2,w
	xorwf	s3,w
	movwf	tmpc

	movf	s0,w
	xorwf	s1,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s0,f

	movf	s1,w
	xorwf	s2,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s1,f

	movf	s2,w
	xorwf	s3,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s2,f

	movf	s3,w
	xorwf	tmp1,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s3,f


	movf	s4,w	; row 2
	movwf	tmp1
	xorwf	s5,w
	xorwf	s6,w
	xorwf	s7,w
	movwf	tmpc

	movf	s4,w
	xorwf	s5,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s4,f

	movf	s5,w
	xorwf	s6,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s5,f

	movf	s6,w
	xorwf	s7,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s6,f

	movf	s7,w
	xorwf	tmp1,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s7,f

	movf	s8,w	; row 3
	movwf	tmp1
	xorwf	s9,w
	xorwf	s10,w
	xorwf	s11,w
	movwf	tmpc

	movf	s8,w
	xorwf	s9,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s8,f

	movf	s9,w
	xorwf	s10,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s9,f

	movf	s10,w
	xorwf	s11,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s10,f

	movf	s11,w
	xorwf	tmp1,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s11,f

	movf	s12,w	; row 4
	movwf	tmp1
	xorwf	s13,w
	xorwf	s14,w
	xorwf	s15,w
	movwf	tmpc

	movf	s12,w
	xorwf	s13,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s12,f

	movf	s13,w
	xorwf	s14,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s13,f

	movf	s14,w
	xorwf	s15,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s14,f

	movf	s15,w
	xorwf	tmp1,w
	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfsc	tmp,7
	xorlw	0x1b
	xorwf	tmpc,w
	xorwf	s15,f

	return
;===============================================================================
; Subroutine to perform mix column function
;===============================================================================
; |                 |    |         |
; | s0  s1  s2  s3  |    | e 9 d b |
; | s4  s5  s6  s7  |    | b e 9 d |
; | s8  s9  s10 s11 |    | d b e 9 |
; | s12 s13 s14 s15 |    | 9 d b e |
; |                 |    |         |

imixcol		xor4	tmp,s0,s1,s2,s3		; row 1
	xor2	tmp1,s0,s2
	get2	tmp1
	xor2	tmp2,s1,s3
	get2	tmp2
	xorwf	tmp1,w
	movwf	tmp3
	get2	tmp3
	get2a	tmp3,tmp
	xor3	d0,s0,s1,tmp1
	get2	d0
	xor3	d1,s1,s2,tmp2
	get2	d1
	xor3	d2,s2,s3,tmp1
	get2	d2
	xor3	tmp1,d0,s0,tmp3
	movwf	s0
	xor3	tmp2,d1,s1,tmp3
	movwf	s1
	xor3	tmp3,d2,s2,tmp3
	movwf	s2
	movf	tmp1,w
	xorwf	tmp2,w
	xorwf	tmp3,w
	xorwf	tmp,w
	movwf	s3
	
	xor4	tmp,s4,s5,s6,s7		; row 2
	xor2	tmp1,s4,s6
	get2	tmp1
	xor2	tmp2,s5,s7
	get2	tmp2
	xorwf	tmp1,w		
	movwf	tmp3
	get2	tmp3
	get2a	tmp3,tmp
	xor3	d0,s4,s5,tmp1
	get2	d0
	xor3	d1,s5,s6,tmp2
	get2	d1
	xor3	d2,s6,s7,tmp1
	get2	d2
	xor3	tmp1,d0,s4,tmp3
	movwf	s4
	xor3	tmp2,d1,s5,tmp3
	movwf	s5
	xor3	tmp3,d2,s6,tmp3
	movwf	s6
	movf	tmp1,w
	xorwf	tmp2,w
	xorwf	tmp3,w
	xorwf	tmp,w
	movwf	s7


	xor4	tmp,s8,s9,s10,s11	; row 3
	xor2	tmp1,s8,s10
	get2	tmp1
	xor2	tmp2,s9,s11
	get2	tmp2
	xorwf	tmp1,w
	movwf	tmp3
	get2	tmp3
	get2a	tmp3,tmp	
	xor3	d0,s8,s9,tmp1
	get2	d0
	xor3	d1,s9,s10,tmp2
	get2	d1
	xor3	d2,s10,s11,tmp1
	get2	d2
	xor3	tmp1,d0,s8,tmp3
	movwf	s8
	xor3	tmp2,d1,s9,tmp3
	movwf	s9
	xor3	tmp3,d2,s10,tmp3
	movwf	s10
	movf	tmp1,w
	xorwf	tmp2,w
	xorwf	tmp3,w
	xorwf	tmp,w
	movwf	s11


	xor4	tmp,s15,s14,s13,s12	; row 4
	xor2	tmp1,s12,s14
	get2	tmp1
	xor2	tmp2,s13,s15
	get2	tmp2
	xorwf	tmp1,w
	movwf	tmp3
	get2	tmp3
	get2a	tmp3,tmp
	xor3	d0,s12,s13,tmp1
	get2	d0
	xor3	d1,s13,s14,tmp2
	get2	d1
	xor3	d2,s14,s15,tmp1
	get2	d2
	xor3	tmp1,d0,s12,tmp3
	movwf	s12
	xor3	tmp2,d1,s13,tmp3
	movwf	s13
	xor3	tmp3,d2,s14,tmp3
	movwf	s14
	movf	tmp1,w
	xorwf	tmp2,w
	xorwf	tmp3,w
	xorwf	tmp,w
	movwf	s15
	return
;===============================================================================
; Sbox and Inverse Sbox Lookup Table
;===============================================================================
lookup	movwf	pcl

        org     ($+0xff)&~0xff  ; force page alignment
sbox	equ ($>>8)
	dt	0x63,0x7c,0x77,0x7b,0xf2,0x6b,0x6f,0xc5,0x30,0x01,0x67,0x2b,0xfe,0xd7,0xab,0x76
	dt	0xca,0x82,0xc9,0x7d,0xfa,0x59,0x47,0xf0,0xad,0xd4,0xa2,0xaf,0x9c,0xa4,0x72,0xc0
	dt	0xb7,0xfd,0x93,0x26,0x36,0x3f,0xf7,0xcc,0x34,0xa5,0xe5,0xf1,0x71,0xd8,0x31,0x15
	dt	0x04,0xc7,0x23,0xc3,0x18,0x96,0x05,0x9a,0x07,0x12,0x80,0xe2,0xeb,0x27,0xb2,0x75
	dt	0x09,0x83,0x2c,0x1a,0x1b,0x6e,0x5a,0xa0,0x52,0x3b,0xd6,0xb3,0x29,0xe3,0x2f,0x84
	dt	0x53,0xd1,0x00,0xed,0x20,0xfc,0xb1,0x5b,0x6a,0xcb,0xbe,0x39,0x4a,0x4c,0x58,0xcf
	dt	0xd0,0xef,0xaa,0xfb,0x43,0x4d,0x33,0x85,0x45,0xf9,0x02,0x7f,0x50,0x3c,0x9f,0xa8
	dt	0x51,0xa3,0x40,0x8f,0x92,0x9d,0x38,0xf5,0xbc,0xb6,0xda,0x21,0x10,0xff,0xf3,0xd2
	dt	0xcd,0x0c,0x13,0xec,0x5f,0x97,0x44,0x17,0xc4,0xa7,0x7e,0x3d,0x64,0x5d,0x19,0x73
	dt	0x60,0x81,0x4f,0xdc,0x22,0x2a,0x90,0x88,0x46,0xee,0xb8,0x14,0xde,0x5e,0x0b,0xdb
	dt	0xe0,0x32,0x3a,0x0a,0x49,0x06,0x24,0x5c,0xc2,0xd3,0xac,0x62,0x91,0x95,0xe4,0x79
	dt	0xe7,0xc8,0x37,0x6d,0x8d,0xd5,0x4e,0xa9,0x6c,0x56,0xf4,0xea,0x65,0x7a,0xae,0x08
	dt	0xba,0x78,0x25,0x2e,0x1c,0xa6,0xb4,0xc6,0xe8,0xdd,0x74,0x1f,0x4b,0xbd,0x8b,0x8a
	dt	0x70,0x3e,0xb5,0x66,0x48,0x03,0xf6,0x0e,0x61,0x35,0x57,0xb9,0x86,0xc1,0x1d,0x9e
	dt	0xe1,0xf8,0x98,0x11,0x69,0xd9,0x8e,0x94,0x9b,0x1e,0x87,0xe9,0xce,0x55,0x28,0xdf
	dt	0x8c,0xa1,0x89,0x0d,0xbf,0xe6,0x42,0x68,0x41,0x99,0x2d,0x0f,0xb0,0x54,0xbb,0x16

isbox	equ ($>>8)
	dt	0x52,0x09,0x6a,0xd5,0x30,0x36,0xa5,0x38,0xbf,0x40,0xa3,0x9e,0x81,0xf3,0xd7,0xfb
	dt	0x7c,0xe3,0x39,0x82,0x9b,0x2f,0xff,0x87,0x34,0x8e,0x43,0x44,0xc4,0xde,0xe9,0xcb
	dt	0x54,0x7b,0x94,0x32,0xa6,0xc2,0x23,0x3d,0xee,0x4c,0x95,0x0b,0x42,0xfa,0xc3,0x4e
	dt	0x08,0x2e,0xa1,0x66,0x28,0xd9,0x24,0xb2,0x76,0x5b,0xa2,0x49,0x6d,0x8b,0xd1,0x25
	dt	0x72,0xf8,0xf6,0x64,0x86,0x68,0x98,0x16,0xd4,0xa4,0x5c,0xcc,0x5d,0x65,0xb6,0x92
	dt	0x6c,0x70,0x48,0x50,0xfd,0xed,0xb9,0xda,0x5e,0x15,0x46,0x57,0xa7,0x8d,0x9d,0x84
	dt	0x90,0xd8,0xab,0x00,0x8c,0xbc,0xd3,0x0a,0xf7,0xe4,0x58,0x05,0xb8,0xb3,0x45,0x06
	dt	0xd0,0x2c,0x1e,0x8f,0xca,0x3f,0x0f,0x02,0xc1,0xaf,0xbd,0x03,0x01,0x13,0x8a,0x6b
	dt	0x3a,0x91,0x11,0x41,0x4f,0x67,0xdc,0xea,0x97,0xf2,0xcf,0xce,0xf0,0xb4,0xe6,0x73
	dt	0x96,0xac,0x74,0x22,0xe7,0xad,0x35,0x85,0xe2,0xf9,0x37,0xe8,0x1c,0x75,0xdf,0x6e
	dt	0x47,0xf1,0x1a,0x71,0x1d,0x29,0xc5,0x89,0x6f,0xb7,0x62,0x0e,0xaa,0x18,0xbe,0x1b
	dt	0xfc,0x56,0x3e,0x4b,0xc6,0xd2,0x79,0x20,0x9a,0xdb,0xc0,0xfe,0x78,0xcd,0x5a,0xf4
	dt	0x1f,0xdd,0xa8,0x33,0x88,0x07,0xc7,0x31,0xb1,0x12,0x10,0x59,0x27,0x80,0xec,0x5f
	dt	0x60,0x51,0x7f,0xa9,0x19,0xb5,0x4a,0x0d,0x2d,0xe5,0x7a,0x9f,0x93,0xc9,0x9c,0xef
	dt	0xa0,0xe0,0x3b,0x4d,0xae,0x2a,0xf5,0xb0,0xc8,0xeb,0xbb,0x3c,0x83,0x53,0x99,0x61
	dt	0x17,0x2b,0x04,0x7e,0xba,0x77,0xd6,0x26,0xe1,0x69,0x14,0x63,0x55,0x21,0x0c,0x7d

	END
